/**
 * OW2 Util
 * Copyright (C) 2008 Bull S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id: WarDeployableMetadataFactory.java 3600 2008-07-04 15:25:28Z gaellalire $
 * --------------------------------------------------------------------------
 */

package org.ow2.util.ee.metadata.war.impl;

import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.ow2.util.archive.api.ArchiveException;
import org.ow2.util.archive.api.IArchive;
import org.ow2.util.ee.deploy.api.deployable.WARDeployable;
import org.ow2.util.ee.deploy.api.deployable.metadata.DeployableMetadataException;
import org.ow2.util.ee.metadata.war.api.IWarDeployableMetadata;
import org.ow2.util.ee.metadata.war.api.IWarDeployableMetadataFactory;
import org.ow2.util.ee.metadata.war.api.exceptions.WARDeploymentDescException;
import org.ow2.util.ee.metadata.war.impl.configurator.WarDeployableMetadataConfigurator;
import org.ow2.util.ee.metadata.war.impl.specific.ISpecificWarDeployableFactory;
import org.ow2.util.ee.metadata.war.impl.xml.Filter;
import org.ow2.util.ee.metadata.war.impl.xml.Listener;
import org.ow2.util.ee.metadata.war.impl.xml.Servlet;
import org.ow2.util.ee.metadata.war.impl.xml.Tag;
import org.ow2.util.ee.metadata.war.impl.xml.WAR;
import org.ow2.util.ee.metadata.war.impl.xml.WARDeploymentDesc;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.scan.api.IClassesLocator;
import org.ow2.util.scan.api.IScanner;
import org.ow2.util.scan.api.ScanException;
import org.ow2.util.scan.api.classlocator.ClassLoaderClassesLocator;
import org.ow2.util.scan.api.configurator.IArchiveConfigurator;
import org.ow2.util.scan.impl.ASMScannerImpl;
import org.ow2.util.url.URLUtils;
import org.ow2.util.url.URLUtilsException;

/**
 * Factory to contruct war deployable metadata.
 * @author Gael Lalire
 */
public class WarDeployableMetadataFactory implements IWarDeployableMetadataFactory {

    /**
     * Logger.
     */
    private static final Log LOGGER = LogFactory.getLog(WarDeployableMetadataFactory.class);

    /**
     * Class scanner.
     */
    private IScanner scanner;

    /**
     * Default constructor. Use {@link ASMScannerImpl} scanner to analyze class bytecode.
     * @throws ScanException never
     */
    public WarDeployableMetadataFactory() throws ScanException {
        this.scanner = new ASMScannerImpl();
    }

    /**
     * Constructor with a scanner.
     * @param scanner the scanner which analyze class bytecode.
     */
    public WarDeployableMetadataFactory(final IScanner scanner) {
        this.scanner = scanner;
    }

    /**
     * Create war deployable metadata.
     * @param deployable the war deployable
     * @return the metadata
     * @throws DeployableMetadataException if error occurs
     */
    public IWarDeployableMetadata createDeployableMetadata(final WARDeployable deployable) throws DeployableMetadataException {
        return createDeployableMetadata(deployable, null, null);
    }

    /**
     * Create war deployable metadata.
     * @param deployable the war deployable
     * @param classLoader a classloader to use if a class is not in archive
     * @return the metadata
     * @throws DeployableMetadataException if error occurs
     */
    public IWarDeployableMetadata createDeployableMetadata(final WARDeployable deployable, final ClassLoader classLoader)
            throws DeployableMetadataException {
        List<? extends IClassesLocator> classesLocatorList = null;
        if (classLoader != null) {
            classesLocatorList = Arrays.asList(new ClassLoaderClassesLocator(classLoader));
        }
        return createDeployableMetadata(deployable, classesLocatorList, null);
    }

    /**
     * Extract a class list from xml.
     * @param war a object representation of web.xml
     * @param classNames a set to fill with class name inside web.xml
     */
    private void completeClassList(final WAR war, final Set<String> classNames) {

        // Add the servlets
        for (Servlet servlet : war.getServlets()) {
            String className = servlet.getClassName();
            if (className != null) {
                classNames.add(className.replace('.', '/'));
            }
        }

        // Add the listeners
        for (Listener listener : war.getListeners()) {
            String className = listener.getListenerClassName();
            if (className != null) {
                classNames.add(className.replace('.', '/'));
            }
        }

        // Add the tags
        for (Tag tag : war.getTags()) {
            String className = tag.getTagClassName();
            if (className != null) {
                classNames.add(className.replace('.', '/'));
            }
        }

        // Add the filters
        for (Filter filter : war.getFilters()) {
            String className = filter.getFilterClassName();
            if (className != null) {
                classNames.add(className.replace('.', '/'));
            }
        }

    }

    /**
     * Create war deployable metadata.
     * @param deployable the war deployable
     * @param warExternLibraries war extern libraries
     * @param specificDeployableFactoryList a specific deployable factory list
     * @return the metadata
     * @throws DeployableMetadataException if error occurs
     */
    public IWarDeployableMetadata createDeployableMetadata(final WARDeployable deployable,
            final List<? extends IClassesLocator> warExternLibraries,
            final List<? extends ISpecificWarDeployableFactory> specificDeployableFactoryList)
            throws DeployableMetadataException {
        try {
            if (!URLUtils.urlToFile2(deployable.getArchive().getURL()).isDirectory()) {
                throw new DeployableMetadataException("war must be uncompressed (local directory expected)");
            }
        } catch (URLUtilsException e) {
            throw new DeployableMetadataException("war must be uncompressed (local directory expected)");
        } catch (ArchiveException e) {
            throw new DeployableMetadataException("war exception", e);
        }

        // for time debugging
        long tAnalyzeStart = System.currentTimeMillis();

        IArchive archive = deployable.getArchive();

        // Get WAR DD.
        WAR war;
        try {
            war = WARDeploymentDesc.getWAR(archive);
        } catch (WARDeploymentDescException e) {
            throw new DeployableMetadataException(e);
        }

        WarDeployableMetadataConfigurator warArchiveMetadataConfigurator = new WarDeployableMetadataConfigurator(deployable);
        Set<String> classNameSet = new HashSet<String>();

        // call before scan on each specific
        if (specificDeployableFactoryList != null) {
            for (ISpecificWarDeployableFactory specificDeployableFactory : specificDeployableFactoryList) {
                warArchiveMetadataConfigurator
                        .addSpecificMetadataConfigurator(specificDeployableFactory.beforeScan(deployable, classNameSet));
            }
        }

        // First step : extract classes from XML
        completeClassList(war, classNameSet);

        // if metadata complete then ignore standard annotation (like @resource)
        if (war.isMetadataComplete()) {
            warArchiveMetadataConfigurator.setAnnotationParsingDesactived(true);
        }

        List<IArchiveConfigurator> archiveConfiguratorList = Arrays.asList((IArchiveConfigurator) warArchiveMetadataConfigurator);

        WarClassesLocator warClassesLocator = new WarClassesLocator(archive);

        // begin scan
        List<String> scannedClasses = new ArrayList<String>();
        List<String> nextClassNameToScan = new ArrayList<String>();
        nextClassNameToScan.addAll(classNameSet);

        try {
            while (nextClassNameToScan.size() != 0) {
                // recalculate classNameSet
                classNameSet.clear();
                for (String className : nextClassNameToScan) {
                    URL url = warClassesLocator.getClassURL(className);
                    if (url == null && warExternLibraries != null) {
                        for (IClassesLocator classesLocator : warExternLibraries) {
                            url = classesLocator.getClassURL(className);
                            if (url != null) {
                                break;
                            }
                        }
                    }
                    if (url != null) {
                        scanner.scanClass(url, classNameSet, archiveConfiguratorList);
                    }
                    // we add to scanned even if there was no scan (class not
                    // found)
                    scannedClasses.add(className);
                }

                // keep what is not scanned
                nextClassNameToScan.clear();
                for (String className : classNameSet) {
                    if (!scannedClasses.contains(className)) {
                        nextClassNameToScan.add(className);
                    }
                }

            }
        } catch (ScanException e) {
            throw new DeployableMetadataException("Error during scan", e);
        }

        // scan finished

        // time if debugging
        if (LOGGER.isDebugEnabled()) {
            long tAnalyzeStartEnd = System.currentTimeMillis();
            if (LOGGER.isDebugEnabled()) {
                LOGGER.debug("Analyze of file {0} took {1} ms.", archive.getName(), (tAnalyzeStartEnd - tAnalyzeStart));
            }
        }

        return warArchiveMetadataConfigurator.getWarDeployableMetadata();
    }

    /**
     * @return the deployable metadata class
     */
    public Class<IWarDeployableMetadata> getDeployableMetadataClass() {
        return IWarDeployableMetadata.class;
    }

}
